package com.fossgalaxy.games.tbs.editor.tools;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.editor.Editor;
import com.fossgalaxy.games.tbs.editor.EditorView.EditorModel;
import com.fossgalaxy.games.tbs.entity.HexUtils;
import com.fossgalaxy.games.tbs.io.map2.MapDef;
import com.fossgalaxy.games.tbs.parameters.EntityType;
import com.fossgalaxy.games.tbs.entity.HexagonTile;
import com.fossgalaxy.games.tbs.parameters.GameSettings;
import com.fossgalaxy.games.tbs.parameters.ResourceType;
import com.fossgalaxy.games.tbs.parameters.TerrainType;
import com.fossgalaxy.games.tbs.ui.UIModel;
import org.codetome.hexameter.core.api.CubeCoordinate;
import org.codetome.hexameter.core.api.Hexagon;
import org.codetome.hexameter.core.api.HexagonalGrid;
import org.codetome.hexameter.core.api.HexagonalGridCalculator;

import javax.swing.*;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Queue;

public class FloodFill implements Tool {
    private EditorModel model;

    public FloodFill(EditorModel model) {
        this.model = model;
    }


    @Override
    public void placeTerrain(MapDef def, CubeCoordinate cube, TerrainType replacement) {

        String targetType = def.getTerrain(cube);
        Queue<CubeCoordinate> queue = new LinkedList<>();

        if ( (replacement == null && targetType == null) || (replacement != null && replacement.getID().equals(targetType)) ) {
            return;
        }

        def.setTerrian(cube, replacement);
        queue.add(cube);

        //FIXME, doing this just to get neighbouring spots is silly.
        GameState state = def.buildState(model.getSettings());

        while (!queue.isEmpty()) {
            CubeCoordinate n = queue.poll();

            Collection<Hexagon<HexagonTile>> neighbors = state.getNeighbors(n);
            for (Hexagon<HexagonTile> hex : neighbors) {

                if (!hex.getSatelliteData().isPresent()) {
                    continue;
                }

                CubeCoordinate neighborPos = hex.getCubeCoordinate();

                String type = def.getTerrain(neighborPos);
                if ( (targetType == null && type == null) || (type != null && type.equals(targetType)) ) {
                    def.setTerrian(hex.getCubeCoordinate(), replacement);
                    queue.add(neighborPos);
                }

            }

        }

    }

    @Override
    public void placeResource(MapDef state, CubeCoordinate cube, ResourceType type) {
        JOptionPane.showMessageDialog(null, "You can't flood fill resources!");
    }

    @Override
    public void placeEntity(MapDef state, CubeCoordinate pos, EntityType selectedEntity, int owner) {
        JOptionPane.showMessageDialog(null, "You can't flood fill entities!");
    }

    @Override
    public String toString() {
        return "Fill";
    }
}
